import dotenv from "dotenv";
import ini from "ini";

import acmeSh from "../../utils/acme.sh";
import vaultAgent from "../../utils/vault-agent";

import { SyncVault } from "./abstracts";

interface AcmeCertificate {
  ca: string;
  fullchain: string;
  cer: string;
  conf: string;
  csr: string;
  csrConf: string;
  key: string;
}

export class CertificateVault extends SyncVault<AcmeCertificate> {
  static basePath = "geektr.co/acme.sh";
  constructor(
    public primaryDomain: string,
    public alternativeDomains?: string[]
  ) {
    super();
  }
  get vaultPath(): string {
    return `${CertificateVault.basePath}/${this.primaryDomain}`;
  }
  static async pullAll(): Promise<CertificateVault[]> {
    const domains = await vaultAgent.list(CertificateVault.basePath);
    return Promise.all(
      domains.map(async (domain) => {
        const vault = new CertificateVault(domain);
        await vault.pull();
        return vault;
      })
    );
  }
  private parseConf() {
    if (!this.data?.conf) return;
    return dotenv.parse(this.data.conf);
  }
  get nextRenewTime(): Date | undefined {
    const conf = this.parseConf();
    if (!conf) return;
    return new Date(conf.Le_NextRenewTimeStr);
  }
  get needsRenew(): boolean {
    const nextRenewTime = this.nextRenewTime;
    return (nextRenewTime && nextRenewTime <= new Date()) || false;
  }
  private parseCsrConf() {
    if (!this.data?.csrConf) return;
    return ini.parse(this.data.csrConf);
  }
  get subjectAltName(): string[] {
    const csrConf = this.parseCsrConf();
    if (!csrConf) return [];
    if (!csrConf.v3_req) throw `'v3_req' not found in csr.conf`;

    // DNS:minio.geektr.co,DNS:*.minio.geektr.co
    const altNames = csrConf.v3_req.subjectAltName as string;
    return altNames.split(",").map((i) => i.slice(4));
  }
  async renew(keepAltStore = false): Promise<void> {
    try {
      await this.load();
      await this.pushAlt();
    } catch (e) {
      if (!(e as { message: string }).message.includes("status code 404"))
        throw e;

      console.log(
        `[acme.sh] ${this.primaryDomain} not found in vault, issue a new cert now...`
      );
    }
    await this.renewAlt();
    await this.pullAlt();
    await this.push();

    if (keepAltStore) return;
    await this.removeAlt();
  }
  // local cert handlers
  async pullAlt(): Promise<void> {
    this.data = await acmeSh.read(this.primaryDomain);
  }
  async pushAlt(): Promise<void> {
    if (!this.data) throw "pull data before push";
    await acmeSh.write(this.primaryDomain, this.data);
  }
  async removeAlt(): Promise<void> {
    await acmeSh.remove(this.primaryDomain);
  }
  async renewAlt(): Promise<void> {
    await acmeSh.autoRenew(
      [
        this.primaryDomain,
        ...(this.subjectAltName || []),
        ...(this.alternativeDomains || []),
      ],
      this.data ? "renew" : "issue"
    );
  }
  static async pullAllAlt(): Promise<CertificateVault[]> {
    const domains = await acmeSh.list();
    return Promise.all(
      domains.map(async (domain) => {
        const vault = new CertificateVault(domain);
        await vault.pullAlt();
        return vault;
      })
    );
  }
}
