import path from "path";

import { Command } from "commander";
import { execFile } from "mz/child_process";

const env = process.env;

function generateDockerLoginCommands() {
  if (!env.CI_REGISTRY_IMAGE) {
    console.error(
      "environment variable $CI_REGISTRY_IMAGE not found, may be you forgot open container registry for current project"
    );
    process.exit(1);
  }

  const registry = env.CI_REGISTRY as string;
  const user = env.CI_REGISTRY_USER as string;
  const password = env.CI_REGISTRY_PASSWORD as string;

  console.log(`docker login -u "${user}" -p "${password}" ${registry}`);
}

async function buildImage() {
  const args: string[] = ["build", "--compress"];

  if (env.file) {
    args.push("--file", env.file);
  }

  const ciDefaultTag = env.CI_REGISTRY_IMAGE;
  if (!ciDefaultTag) {
    console.error(
      "environment variable $CI_REGISTRY_IMAGE not found, may be you forgot open container registry for current project"
    );
    process.exit(1);
  }

  const tagPath = env.sub_path
    ? path.join(ciDefaultTag, env.sub_path)
    : ciDefaultTag;

  const tags = [];

  if (typeof env.tags === "string") {
    env.tags.split(",").forEach((i) => tags.push(i));
  } else if (typeof env.tag === "string" && env.tag.length) {
    tags.push(env.tag);
  } else {
    tags.push("latest");
  }

  const useHashTag = env.use_hash_tag === "true";
  if (useHashTag) {
    tags.push(env.CI_COMMIT_SHA as string);
  }

  const completeTags = tags.map((i) => `${tagPath}:${i}`);
  const completeCacheFrom = `${tagPath}:${env.cache_from || tags[0]}`;
  const noCache = env.no_cache === "true";

  if (noCache) {
    args.push("--no-cache");
  } else {
    args.push("--cache-from", completeCacheFrom);
  }

  completeTags.forEach((tag) => args.push("--tag", tag));

  args.push(env.workdir ? env.workdir : ".");

  const buildArgs = (env.build_args || "").match(/(\\.|[^,])+/g);
  if (buildArgs) {
    buildArgs.forEach((buildArg) => args.push("--build-arg", buildArg));
  }

  try {
    console.warn(`docker ${args.join(" ")}`);
    await execFile("docker", args);
    console.warn("Build Complete");

    for (const tag of completeTags) {
      console.warn(`docker push ${tag}`);
      await execFile("docker", ["push", tag]);
    }
    console.warn("Push Complete");

    for (const tag of completeTags) {
      console.warn(`docker rmi ${tag}`);
      await execFile("docker", ["rmi", tag]);
    }
    console.warn("Local Image Remove Complete");
  } catch (e) {
    console.error(e);
    process.exit(1);
  }
}

export default function (program: Command): void {
  const dockerCi = program
    .command("docker-ci")
    .description("docker ci helpers");

  dockerCi
    .command("login")
    .description("generate docker login command")
    .on("--help", () =>
      console.warn(`
Usage:
  eval \`yumemi docker-ci login\``)
    )
    .action(generateDockerLoginCommands);

  dockerCi
    .command("build")
    .description("build image by provided environment variables for gitlab ci")
    .on("--help", () =>
      console.warn(`
Environemt Variables:
  workdir:      Work direcotry
  file:         Name of the Dockerfile
  sub_path:     Sub path for registy name
  tag:          Image Tag
  tags:         Image Tags
  no_cache:     No Cache
  cache_from:   Cache From
  use_hash_tag: Use git hash as tag
  build_args:   Set build-time variables, format: 'KEY=VALUE,KEY2=VALUE2'`)
    )
    .action(buildImage);
}
