import os from "os";
import path from "path";

import commandExists from "command-exists";
import { Command } from "commander";
import inquirer from "inquirer";
import { blue, green, red, underline, yellow } from "kleur";
import { execFile } from "mz/child_process";
import fs from "mz/fs";

import AssemblyLine from "../../utils/assembly-line";
import TreeView from "../../utils/tree-view";

import Node from "./node";

async function getProjjBaseDirectory(): Promise<string | undefined> {
  const configFile = path.resolve(os.homedir(), ".projj/config.json");
  try {
    const config = JSON.parse((await fs.readFile(configFile)).toString());
    if (config.base) {
      return config.base;
    }
    // eslint-disable-next-line no-empty
  } catch {}
}

class ScanAssemblyLine extends AssemblyLine<Node> {
  loadGitStatus?: boolean;
  constructor(rootNode: Node, loadGitStatus = false) {
    super({ artifacts: [rootNode] });
    this.loadGitStatus = loadGitStatus;
  }
  async work(node: Node): Promise<void> {
    await node.load({ gitStatus: this.loadGitStatus, description: true });
    if (node.isDirectory && !node.isProject && node.children) {
      this.push(node.children);
    }
  }
}

class ProjectsTreeView extends TreeView<Node> {
  renderNode(node: Node): string {
    const basename = path.basename(node.path);
    if (!node.readable) return `${red(basename)} [permission denied]`;
    if (!node.isDirectory) return `${yellow(basename)}`;
    if (node.isProject) {
      const comps = [blue().bold(basename)];
      if (node.description) comps.push(green(node.description));
      if (node.gitStatus) {
        const git = node.gitStatus;
        const gitStatusText = [
          git.current + (git.isClean() ? "" : "*"),
          git.ahead && `${git.ahead}↑`,
          git.behind && `${git.ahead}↓`,
        ]
          .filter((i) => i)
          .join(" ");

        comps.push(git.isClean() ? gitStatusText : underline(gitStatusText));
      }
      return comps.join(" | ");
    }
    return basename;
  }
}

async function getVsCodeCommand(): Promise<string> {
  try {
    return await commandExists("code");
    // eslint-disable-next-line no-empty
  } catch {}
  try {
    return await commandExists("code-insiders");
    // eslint-disable-next-line no-empty
  } catch {}
  throw Error("neither 'code' nor 'code-insiders' are in path");
}

export default function (program: Command): void {
  program
    .command("projt [target directory]")
    .description("show all projects in tree view")
    .option("-d, --depth [number]", "max search depth", "6")
    .option("--git", "show git status for each project")
    .option("--select", "select item and echo path")
    .option("--vscode", "open project with vscode")
    .action(
      async (
        targetDir,
        {
          depth: maxDepth,
          git: loadGitStatus = false,
          select: showSelector = false,
          vscode: openWithVsCode = false,
        }
      ) => {
        maxDepth = Number(maxDepth);
        targetDir = path.resolve(
          targetDir || (await getProjjBaseDirectory()) || "."
        );
        const rootNode = new Node(targetDir);
        if (openWithVsCode) {
          showSelector = true;
        }

        await new ScanAssemblyLine(rootNode, loadGitStatus).run();
        const lines = new ProjectsTreeView(rootNode).render(maxDepth);

        if (!showSelector) {
          console.log(lines.map((i) => i.boxes.join("") + i.view).join("\n"));
          return;
        }

        const selectedNode: Node = (
          await inquirer.prompt([
            {
              type: "list",
              name: "selectedNode",
              choices: lines.map((i) => ({
                name: i.boxes.join("") + i.view,
                value: i.node,
              })),
            },
          ])
        ).selectedNode;

        if (!openWithVsCode) {
          console.log(selectedNode.path);
          return;
        }

        if (openWithVsCode) {
          const vsCodeCommand = await getVsCodeCommand();
          await execFile(vsCodeCommand, [selectedNode.path]);
        }
      }
    );
}
