import path from "path";

import { LazyGetter } from "lazy-get-decorator";
import fs, { Stats } from "mz/fs";
import SimpleGit from "simple-git/promise";

const projectDetermineBasis = [
  ".git",
  "package.json",
  "yarn.lock",
  "node_modules",
  "Dockerfile",
  "docker-compose.yml",
  "docker-compose.yaml",
];

interface NodeLoadOptions {
  gitStatus?: boolean;
  description?: boolean;
}

export default class Node {
  path: string;

  readable = false;

  stats?: Stats;

  isProject?: boolean;
  subDirs?: string[];
  description?: string;
  gitStatus?: SimpleGit.StatusResult;

  constructor(path: string) {
    this.path = path;
  }

  async load(opts?: NodeLoadOptions): Promise<void> {
    await this.updateReadable();
    if (!this.readable) return;

    await this.updateStats();

    if (this.isDirectory) {
      if (!(await this.updateSubDirs()).length) return;
      if (!(await this.updateIsProject())) return;
      if (opts?.description) await this.updateDescription();
      if (opts?.gitStatus) await this.updateGitStatus();
    }
  }

  public async updateReadable(): Promise<boolean> {
    try {
      fs.access(this.path, fs.constants.R_OK);
      return (this.readable = true);
    } catch {
      return (this.readable = false);
    }
  }

  async updateSubDirs(): Promise<string[]> {
    return (this.subDirs = await fs.readdir(this.path));
  }

  async updateDescription(): Promise<string | void> {
    const readSubFile = async (file: string): Promise<string> =>
      (await fs.readFile(path.join(this.path, file))).toString();

    const subDirs = this.subDirs;
    if (!subDirs) throw Error("required property: subDirs");

    if (subDirs.includes("package.json")) {
      return (this.description = JSON.parse(
        await readSubFile("package.json")
      ).description);
    }

    const readme = subDirs.find((i) => i.toLowerCase() === "readme.md");
    if (readme) {
      const match = (await readSubFile(readme)).match(/# (.+)/);
      if (match) return (this.description = match[1]);
    }
  }

  async updateIsProject(): Promise<boolean> {
    if (!this.subDirs) throw Error("required property: subDirs");
    return (this.isProject = !!this.subDirs.find((i) =>
      projectDetermineBasis.includes(i)
    ));
  }

  async updateGitStatus(): Promise<SimpleGit.StatusResult | void> {
    try {
      return (this.gitStatus = await SimpleGit(this.path).status());
      // eslint-disable-next-line no-empty
    } catch {}
  }

  async updateStats(): Promise<Stats> {
    return (this.stats = await fs.stat(this.path));
  }

  get isDirectory(): boolean {
    return !!this.stats?.isDirectory();
  }

  get isFile(): boolean {
    return !!this.stats?.isDirectory();
  }

  @LazyGetter()
  get children(): Node[] | undefined {
    if (!this.subDirs) return;
    if (this.isProject) return;
    return this.subDirs.map((name) => new Node(path.join(this.path, name)));
  }
}
