import { promises as fs } from "fs";

import { exec, ExecOptions } from "mz/child_process";

enum OpCategory {
  Login = 1,
  CreditCard = 2,
  SecureNote = 3,
  Identity = 4,
  Password = 5,
  Document = 6,
  SoftwareLicense = 100,
  BankAccount = 101,
  Database = 102,
  DriverLicense = 103,
  OutdoorLicense = 104,
  Membership = 105,
  Passport = 106,
  RewardProgram = 107,
  SocialSecurityNumber = 108,
  WirelessRouter = 109,
  Server = 110,
  EmailAccount = 111,
}

class OpError extends Error {
  constructor(message: string) {
    super(`[1Password] ${message}`);
  }
}

process.on("unhandledRejection", (e) => {
  if (!(e instanceof OpError)) throw e;
  console.error(OpSessionExpired.message);
  process.exit(1);
});

const OpSessionExpired = new OpError(
  "session expired, please refresh by:\nop_login reload"
);
const OpReadSessionFailed = new OpError(
  "failed to read session file, please refresh by:\nop_login reload"
);

export const readOpSession = async (): Promise<string> => {
  const cacheFile = process.env.OP_CACHE_FILE;
  if (!cacheFile) throw OpReadSessionFailed;
  try {
    const session = await fs.readFile(cacheFile, "utf8");
    return session.trim();
  } catch (e) {
    throw OpReadSessionFailed;
  }
};

export const opExec = async (
  command: string,
  options?: ({ encoding?: BufferEncoding } & ExecOptions) | null
): Promise<OpRawData> => {
  const OP_SESSION_my = process.env.OP_SESSION_my || (await readOpSession());

  try {
    const [result] = await exec(command, {
      ...options,
      env: { ...process.env, OP_SESSION_my, ...options?.env },
    });
    return JSON.parse(result) as OpRawData;
  } catch (e) {
    if ((e as { message: string }).message.includes("session expired"))
      throw OpSessionExpired;
    throw e;
  }
};

export interface OnePasswordGetOptions {
  vault?: string;
  item: string;
  path?: string;
}

export interface OpRawData {
  templateUuid: string;
  overview: {
    title: string;
  };
  details: {
    password: string;
    fields?: { id: "username" | "password"; value: string }[];
    sections: {
      fields: { k: string; t: string; v: string }[];
      title: string;
    }[];
  };
}

export interface OpItem {
  title: string;
  type: OpCategory;
  fields: Record<string, string>;
  sections: Record<string, Record<string, string>>;
  _raw: OpRawData;
}

function parseOptions(data: OpRawData): OpItem {
  const title = data.overview.title;
  const type = Number(data.templateUuid);

  const { details } = data;

  const fields: Record<string, string> = {};
  details.fields &&
    details.fields.forEach((f) => {
      if (!f.id) return;
      fields[f.id] = f.value;
    });
  const sections: Record<string, Record<string, string>> = {};
  details.sections &&
    details.sections.forEach((s) => {
      if (!s.fields) return;
      const sectionFields: Record<string, string> = {};
      s.fields.forEach((f) => (sectionFields[f.t] = f.v));
      sections[s.title] = sectionFields;
    });

  return { title, type, fields, sections, _raw: data };
}

export const opGet = async (opts: OnePasswordGetOptions): Promise<OpItem> => {
  let command = "";
  command += `op get item '${opts.item}' `;
  if (opts.vault) command += `--vault '${opts.vault}'`;
  const result = await opExec(command);
  return parseOptions(result);
};
