import { promises as dns } from "dns";
import { promises as fs } from "fs";
import path from "path";

import { ExecFileOptions, spawn } from "mz/child_process";
import psl from "psl";

import { MyAliyunRamsVault } from "../core/vaults/aliyun-ram";
import { MyCloudflareVault } from "../core/vaults/cloudflare";

interface AcmeCertificate {
  ca: string;
  fullchain: string;
  cer: string;
  conf: string;
  csr: string;
  csrConf: string;
  key: string;
}

interface DnsProviderConfig {
  // fs.readdirSync(`${process.env.LE_WORKING_DIR}/dnsapi`).filter(i=>i!='README.md').map(i=>i.slice(0,-3))
  name: string;
  domains: string[];
  providerAsseccKey: () => Promise<unknown>;
}

const DnsProviders: DnsProviderConfig[] = [
  {
    name: "dns_ali",
    domains: ["hichina.com", "alidns.com"],
    providerAsseccKey: () => MyAliyunRamsVault.load(),
  },
  {
    name: "dns_cf",
    domains: ["cloudflare.com"],
    providerAsseccKey: () => MyCloudflareVault.load(),
  },
];

const basePath = process.env.LE_WORKING_DIR || "";

async function handerEachFile(
  domain: string,
  handler: (key: keyof AcmeCertificate, file: string) => Promise<string | void>
): Promise<AcmeCertificate> {
  const entries: [keyof AcmeCertificate, string][] = [
    ["ca", "ca.cer"],
    ["fullchain", `fullchain.cer`],
    ["cer", `${domain}.cer`],
    ["conf", `${domain}.conf`],
    ["csr", `${domain}.csr`],
    ["csrConf", `${domain}.csr.conf`],
    ["key", `${domain}.key`],
  ];

  return Object.fromEntries(
    await Promise.all(
      entries.map(async ([key, file]) => [
        key,
        (await handler(key, file)) || "",
      ])
    )
  );
}

function acmeResolve(subPath: string) {
  if (!basePath) throw `environment variable "$LE_WORKING_DIR" not found.`;
  return path.resolve(basePath, subPath);
}

function acmeExec(args: string[], _opts?: ExecFileOptions) {
  const opts: ExecFileOptions = {
    ..._opts,
    env: {
      ...process.env,
      ..._opts?.env,
    },
  };
  const child = spawn(acmeResolve("acme.sh"), args, opts);
  child.stdout.pipe(process.stdout);
  child.stderr.pipe(process.stderr);
  return new Promise((resolve, reject) => {
    child.on("close", resolve);
    child.on("error", reject);
  });
}

function getMainDomain(domain: string) {
  const parsed = psl.parse(domain);
  if (parsed.error) throw `domain ${domain} invalid: ${parsed.error.code}`;
  if (!parsed.domain) throw `domain ${domain} not includes main domain`;
  return parsed.domain;
}

async function getDNSProvider(domain: string) {
  const providerDomains = await dns.resolveNs(getMainDomain(domain));
  const providerMainDomain = getMainDomain(providerDomains[0]);
  const provider = DnsProviders.find((i) =>
    i.domains.includes(providerMainDomain)
  );
  return provider;
}

async function exist(domain: string): Promise<boolean> {
  try {
    await fs.stat(acmeResolve(domain));
    return true;
  } catch (e) {
    return false;
  }
}

async function ensure(domain: string): Promise<void> {
  if (await exist(domain)) return;
  await fs.mkdir(acmeResolve(domain));
}

async function read(domain: string): Promise<AcmeCertificate> {
  if (!(await exist(domain)))
    throw `[acme.sh] files about domain "${domain}" not found`;

  return await handerEachFile(
    domain,
    async (_, file) =>
      await fs.readFile(acmeResolve(path.join(domain, file)), "utf8")
  );
}

async function write(domain: string, cert: AcmeCertificate): Promise<void> {
  await ensure(domain);

  await handerEachFile(
    domain,
    async (key, file) =>
      await fs.writeFile(
        acmeResolve(path.join(domain, file)),
        cert[key],
        "utf8"
      )
  );
}

async function remove(domain: string): Promise<void> {
  if (!(await exist(domain)))
    throw `[acme.sh] files about domain "${domain}" not found`;

  await handerEachFile(
    domain,
    async (_, file) => await fs.unlink(acmeResolve(path.join(domain, file)))
  );

  await fs.rmdir(acmeResolve(domain));
}

async function autoRenew(
  [primaryDomain, ...alternativeDomains]: string[],
  type: "renew" | "issue"
): Promise<void> {
  const provider = await getDNSProvider(primaryDomain);
  if (!provider) throw `unknown dns provider for domain "${primaryDomain}"`;

  const args: string[] = [];

  args.push(`--${type}`);

  args.push("--dns", provider.name);

  const domainList = [
    primaryDomain,
    ...new Set(alternativeDomains.filter((i) => i !== primaryDomain).sort()),
  ];
  for (const domain of domainList) {
    args.push("-d", domain);
  }

  console.warn(`[acme.sh] ${type} domain ${domainList.join(",")}`);

  const env = (await provider.providerAsseccKey()) as NodeJS.ProcessEnv;
  await acmeExec(args, { env });
}

async function list(): Promise<string[]> {
  const dirs = await fs.readdir(basePath);
  const domains: (string | null)[] = await Promise.all(
    dirs.map(async (domain) => {
      try {
        await fs.stat(path.resolve(basePath, domain, `${domain}.csr`));
        return domain;
      } catch {
        return null;
      }
    })
  );
  return domains.filter((i) => i !== null) as string[];
}

const acmeSh = {
  read,
  write,
  remove,
  autoRenew,
  list,
};

export default acmeSh;
