import os from "os";

import axios from "axios";
import prettyDuration from "pretty-ms";

import { opGet } from "../../utils/1password";
import debouncePromise from "../../utils/debounce-promise";
import JsonDB, { JsonDBFile } from "../../utils/jsondb";

import { baseURL } from "./vault-axios";

let ShouldLogExpire = true;

export interface CreateTokenResponse {
  request_id: string;
  lease_id: string;
  renewable: boolean;
  lease_duration: number;
  data?: unknown;
  wrap_info?: unknown;
  warnings?: unknown;
  auth: {
    client_token: string;
    accessor: string;
    policies: string[];
    token_policies: string[];
    metadata?: unknown;
    lease_duration: number;
    renewable: boolean;
    entity_id: string;
    token_type: string;
    orphan: boolean;
  };
}

async function getTemporaryToken(): Promise<CreateTokenResponse> {
  const opItem = await opGet({
    vault: "geektr.co",
    item: "vault/root",
  });
  const token = opItem._raw.details.password;

  const { data } = await axios.post<CreateTokenResponse>(
    `${baseURL}/auth/token/create`,
    {
      explicit_max_ttl: "1h",
      display_name: `${os.userInfo().username}@${os.hostname()}`,
    },
    { headers: { "X-Vault-Token": token } }
  );
  return data;
}

@JsonDBFile("vault-token.json")
class VaultTokenDB extends JsonDB {
  expire = 0;
  token = "";
  constructor() {
    super();
    this.load();
  }
  get timeLeft() {
    return this.expire - new Date().getTime();
  }
  private _expireLog(shouldLog = ShouldLogExpire) {
    if (shouldLog) {
      console.warn(
        `[Vault] Token will expire ${prettyDuration(this.timeLeft, {
          compact: true,
        })}`
      );
      ShouldLogExpire = false;
    }
  }
  @debouncePromise({ debounceLevel: "class" })
  async initialize() {
    if (this.token && this.timeLeft > 0) {
      this._expireLog();
      return;
    }

    console.warn("[Vault] Token Refresh...");

    const data = await getTemporaryToken();

    this.token = data.auth.client_token;

    this.expire = new Date().getTime() + 1000 * data.auth.lease_duration;
    this.save();
    this._expireLog(true);
  }
}

export default new VaultTokenDB();
